#include <owl/applicat.h>
#include <owl/framewin.h>

#include <gl/gl.h>
#include <gl/glu.h>


unsigned char oneto8[2] = {
  0, 255
};

unsigned char twoto8[4] = {
  0, 0x55, 0xaa, 0xff
};

unsigned char threeto8[8] = {
  0, 36, 73, 109, 146, 182, 219, 255
};

int defaultOverride[13] = {
  0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

PALETTEENTRY defaultPalEntry[20] = {
  { 0,    0,    0,    0 },
  { 0x80, 0,    0,    0 },
  { 0,    0x80, 0,    0 },
  { 0x80, 0x80, 0,    0 },
  { 0,    0,    0x80, 0 },
  { 0x80, 0,    0x80, 0 },
  { 0,    0x80, 0x80, 0 },
  { 0xC0, 0xC0, 0xC0, 0 },

  { 192,  220,  192,  0 },
  { 166,  202,  240,  0 },
  { 255,  251,  240,  0 },
  { 160,  160,  164,  0 },

  { 0x80, 0x80, 0x80, 0 },
  { 0xFF, 0,    0,    0 },
  { 0,    0xFF, 0,    0 },
  { 0xFF, 0xFF, 0,    0 },
  { 0,    0,    0xFF, 0 },
  { 0xFF, 0,    0xFF, 0 },
  { 0,    0xFF, 0xFF, 0 },
  { 0xFF, 0xFF, 0xFF, 0 }
};

//
// class TTestWin
// ~~~~~ ~~~~~~~~
//
class TTestWin : public TWindow {
  public:
    TTestWin(TWindow* parent);

    void SetupWindow  ();
    void CleanupWindow();

    void Paint(TDC& dc, bool erase, TRect& rect);

  protected:
    bool EvEraseBkgnd(HDC hDC);

    void EvPaletteChanged(HWND hWndPalChg);

    bool EvQueryNewPalette();

    void EvSize(uint sizeType, TSize& size);

  private:
    TClientDC* dc;

    TPalette* palette;

    TSize oldSize;

    void InitializeOpenGL();
    bool SetupPixelFormat();
    void SetupRGBPalette ();

    void DrawScene();

    unsigned char ComponentFromIndex(int i, unsigned nbits, unsigned shift);

  DECLARE_RESPONSE_TABLE(TTestWin);
};

DEFINE_RESPONSE_TABLE1(TTestWin, TWindow)
  EV_WM_ERASEBKGND,
  EV_WM_PALETTECHANGED,
  EV_WM_QUERYNEWPALETTE,
  EV_WM_SIZE,
END_RESPONSE_TABLE;

TTestWin::TTestWin(TWindow* parent)
:
 TWindow(parent),
 dc(0), palette(0), oldSize(0, 0)
{
  Attr.Style |= WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
}

void
TTestWin::SetupWindow()
{
  TWindow::SetupWindow();

  dc = new TClientDC(*this);

  // Initialize OpenGL rendering
  //
  InitializeOpenGL();
}

void
TTestWin::CleanupWindow()
{
  wglMakeCurrent(0, 0);

  HGLRC hrc = wglGetCurrentContext();
  if (hrc)
    wglDeleteContext(hrc);

  if (palette)
    dc->RestorePalette();

  delete palette;
  
  delete dc;

  TWindow::CleanupWindow();
}

void
TTestWin::Paint(TDC& /* dc */, bool /* erase */, TRect& /* rect */)
{
  DrawScene();
}

bool
TTestWin::EvEraseBkgnd(HDC /* hDC */)
{
  return true;
}

void
TTestWin::EvPaletteChanged(HWND hWndPalChg)
{
  if (palette && hWndPalChg != *this) {
    dc->RestorePalette();

    dc->SelectObject  (*palette);
    dc->RealizePalette();

    dc->UpdateColors();
  }
}

bool
TTestWin::EvQueryNewPalette()
{
  if (palette) {
    dc->RestorePalette();

    dc->SelectObject  (*palette);
    dc->RealizePalette();

    Invalidate();

    return true;
  }

  return false;
}

void
TTestWin::EvSize(uint sizeType, TSize& size)
{
  TWindow::EvSize(sizeType, size);

  if (size.cy > 0) {
    glViewport(0, 0, size.cx, size.cy);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (GLfloat)size.cx / size.cy, 0.1f, 100.0f);
    glMatrixMode(GL_MODELVIEW);

    if (size.cx < oldSize.cx || size.cy < oldSize.cy)
      DrawScene();

    oldSize = size;
  }
}

void
TTestWin::InitializeOpenGL()
{
  if (!SetupPixelFormat())
    return;

  SetupRGBPalette();

  HGLRC hrc = wglCreateContext(*dc);
  wglMakeCurrent(*dc, hrc);

  glShadeModel(GL_SMOOTH);                            // Enable smooth shading

  glClearDepth(1.0f);                                 // Depth buffer setup

  glEnable(GL_DEPTH_TEST);                            // Enalbe depth testing

  glDepthFunc(GL_LEQUAL);                             // Type of depth test to do

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);  // Really nice perspective calculation
}

bool
TTestWin::SetupPixelFormat()
{
  PIXELFORMATDESCRIPTOR pfd = {
    sizeof(PIXELFORMATDESCRIPTOR),  // size
    1,                              // version
    PFD_DRAW_TO_WINDOW |            // support window
    PFD_SUPPORT_OPENGL |            // support OpenGL
    PFD_DOUBLEBUFFER,               // double buffered
    PFD_TYPE_RGBA,                  // RGBA type
    24,                             // 16-bit color depth
    0, 0, 0, 0, 0, 0,               // color bits ignored
    0,                              // no alpha buffer
    0,                              // alpha bits ignored
    0,                              // no accumulation buffer
    0, 0, 0, 0,                     // accum bits ignored
    32,                             // 16-bit z-buffer
    0,                              // no stencil buffer
    0,                              // no auxiliary buffer
    PFD_MAIN_PLANE,                 // main layer
    0,                              // reserved
    0, 0, 0,                        // layer masks ignored
  };

  int pixelFormat = ChoosePixelFormat(*dc, &pfd);
  if (pixelFormat == 0) {
    MessageBox("ChoosePixelFormat failed.", "Error", MB_ICONERROR | MB_OK);

    return false;
  }

  if (!SetPixelFormat(*dc, pixelFormat, &pfd)) {
    MessageBox("SetPixelFormat failed.", "Error", MB_ICONERROR | MB_OK);

    return false;
  }

  return true;
}

void
TTestWin::SetupRGBPalette()
{
  int pixelFormat = GetPixelFormat(*dc);

  PIXELFORMATDESCRIPTOR pfd;
  DescribePixelFormat(*dc, pixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

  if (pfd.dwFlags & PFD_NEED_PALETTE) {
    int paletteSize = 1 << pfd.cColorBits;

    LOGPALETTE* pPal = (LOGPALETTE*)
        new char[sizeof(LOGPALETTE) + (paletteSize - 1) * sizeof(PALETTEENTRY)];

    pPal->palVersion    = 0x300;
    pPal->palNumEntries = (WORD)paletteSize;

    for (int i = 0; i < paletteSize; ++i) {
      pPal->palPalEntry[i].peRed   =
                         ComponentFromIndex(i, pfd.cRedBits,   pfd.cRedShift);
      pPal->palPalEntry[i].peGreen =
                         ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
      pPal->palPalEntry[i].peBlue  =
                         ComponentFromIndex(i, pfd.cBlueBits,  pfd.cBlueShift);
      pPal->palPalEntry[i].peFlags = 0;
    }

    // Fix up the palette to include the default GDI palette
    //
    if (pfd.cColorBits == 8                         &&
        pfd.cRedBits   == 3 && pfd.cRedShift   == 0 &&
        pfd.cGreenBits == 3 && pfd.cGreenShift == 3 &&
        pfd.cBlueBits  == 2 && pfd.cBlueShift  == 6)
      for (int i = 1 ; i <= 12 ; i++)
        pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];

    palette = new TPalette(pPal);

    if (palette) {
      dc->SelectObject  (*palette);
      dc->RealizePalette();
    }

    delete[] pPal;
  }
}

void
TTestWin::DrawScene()
{
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);               // Black background

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // Clear the screen and the depth buffer

  glLoadIdentity();                                   // Reset the view

  glTranslatef(-1.5f, 0.0f, -6.0f);                   // Move left 1.5 units and into the screen 6 units

  glBegin(GL_TRIANGLES);                              // Draw a triangle
    glColor3f ( 1.0f, 0.0f, 0.0f);                    // Set the color to red
    glVertex3f( 0.0f,  1.0f, 0.0f);                   // Top

    glColor3f ( 0.0f,  1.0f, 0.0f);                   // Set the color to green
    glVertex3f(-1.0f, -1.0f, 0.0f);                   // Bottom left

    glColor3f ( 0.0f,  0.0f, 1.0f);                   // Set the color to blue
    glVertex3f( 1.0f, -1.0f, 0.0f);                   // Bottom right
  glEnd();                                            // Finish drawing the triangle

  glTranslatef(3.0f, 0.0, 0.0);                       // Move right 3.0 units

  glColor3f(0.5f, 0.5f, 1.0f);                        // Set the color to blue one time only
  glBegin(GL_QUADS);                                  // Draw a quad
    glVertex3f(-1.0f,  1.0f, 0.0f);                   // Top left
    glVertex3f( 1.0f,  1.0f, 0.0f);                   // Top right
    glVertex3f( 1.0f, -1.0f, 0.0f);                   // Bottom right
    glVertex3f(-1.0f, -1.0f, 0.0f);                   // Bottom left
  glEnd();                                            // Finish drawing the quad

  SwapBuffers(*dc);
}

unsigned char
TTestWin::ComponentFromIndex(int i, unsigned nbits, unsigned shift)
{
  unsigned char val = (unsigned char)(i >> shift);

  switch (nbits) {
    case 1:
      val &= 0x1;
      return oneto8[val];

    case 2:
      val &= 0x3;
      return twoto8[val];

    case 3:
      val &= 0x7;
      return threeto8[val];

    default:
      return 0;
  }
}

//
// class TTestApp
// ~~~~~ ~~~~~~~~
//
class TTestApp : public TApplication {
  public:
    TTestApp(const char* name);

    void InitMainWindow();
};

TTestApp::TTestApp(const char* name)
:
 TApplication(name)
{
}

void
TTestApp::InitMainWindow()
{
  TFrameWindow* frame = new TFrameWindow(0, "OpenGL Test", new TTestWin(0));
  SetMainWindow(frame);
}

int
OwlMain(int /* argc */, char* /* argv */ [])
{
  return TTestApp("OpenGL Application").Run();
}